@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2008-2010
+Copyright (c) 2008-2011
 	Lars-Dominik Braun <lars@6xq.net>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -21,7 +21,10 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */
 
+#ifndef __FreeBSD__
 #define _BSD_SOURCE /* required by strdup() */
+#define _DARWIN_C_SOURCE /* strdup() on OS X */
+#endif
 
 #include <stdio.h>
 #include <string.h>
@@ -39,7 +42,7 @@ THE SOFTWARE.
 #include "crypt.h"
 #include "config.h"
 
-#define PIANO_PROTOCOL_VERSION "29"
+#define PIANO_PROTOCOL_VERSION "32"
 #define PIANO_RPC_HOST "www.pandora.com"
 #define PIANO_RPC_PORT "80"
 #define PIANO_RPC_PATH "/radio/xmlrpc/v" PIANO_PROTOCOL_VERSION "?"
@@ -57,32 +60,29 @@ void PianoInit (PianoHandle_t *ph) {
 			(unsigned long) time (NULL) % 10000000);
 }
 
+/*	destroy artist linked list
+ */
+void PianoDestroyArtists (PianoArtist_t *artists) {
+	PianoArtist_t *curArtist, *lastArtist;
+
+	curArtist = artists;
+	while (curArtist != NULL) {
+		free (curArtist->name);
+		free (curArtist->musicId);
+		free (curArtist->seedId);
+		lastArtist = curArtist;
+		curArtist = curArtist->next;
+		free (lastArtist);
+	}
+}
+
 /*	free complete search result
  *	@public yes
  *	@param search result
  */
 void PianoDestroySearchResult (PianoSearchResult_t *searchResult) {
-	PianoArtist_t *curArtist, *lastArtist;
-	PianoSong_t *curSong, *lastSong;
-
-	curArtist = searchResult->artists;
-	while (curArtist != NULL) {
-		free (curArtist->name);
-		free (curArtist->musicId);
-		lastArtist = curArtist;
-		curArtist = curArtist->next;
-		free (lastArtist);
-	}
-
-	curSong = searchResult->songs;
-	while (curSong != NULL) {
-		free (curSong->title);
-		free (curSong->artist);
-		free (curSong->musicId);
-		lastSong = curSong;
-		curSong = curSong->next;
-		free (lastSong);
-	}
+	PianoDestroyArtists (searchResult->artists);
+	PianoDestroyPlaylist (searchResult->songs);
 }
 
 /*	free single station
@@ -128,12 +128,21 @@ void PianoDestroyPlaylist (PianoSong_t *playlist) {
 		free (curSong->stationId);
 		free (curSong->album);
 		free (curSong->artistMusicId);
+		free (curSong->feedbackId);
+		free (curSong->seedId);
+		free (curSong->detailUrl);
 		lastSong = curSong;
 		curSong = curSong->next;
 		free (lastSong);
 	}
 }
 
+void PianoDestroyStationInfo (PianoStationInfo_t *info) {
+	PianoDestroyPlaylist (info->feedback);
+	PianoDestroyPlaylist (info->songSeeds);
+	PianoDestroyArtists (info->artistSeeds);
+}
+
 /*	destroy genre linked list
  */
 void PianoDestroyGenres (PianoGenre_t *genres) {
@@ -154,7 +163,6 @@ void PianoDestroyGenres (PianoGenre_t *genres) {
 void PianoDestroyUserInfo (PianoUserInfo_t *user) {
 	free (user->webAuthToken);
 	free (user->authToken);
-	free (user->listenerId);
 }
 
 /*	frees the whole piano handle structure
@@ -242,7 +250,16 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 							"rid=%s&method=sync", ph->routeId);
 					break;
 
-				case 1:
+				case 1: {
+					char *xmlencodedPassword = NULL;
+
+					/* username == email address does not contain &,<,>," */
+					if ((xmlencodedPassword =
+							PianoXmlEncodeString (logindata->password)) ==
+							NULL) {
+						return PIANO_RET_OUT_OF_MEMORY;
+					}
+
 					snprintf (xmlSendBuf, sizeof (xmlSendBuf), 
 							"<?xml version=\"1.0\"?><methodCall>"
 							"<methodName>listener.authenticateListener</methodName>"
@@ -250,18 +267,19 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 							"<param><value><string>%s</string></value></param>"
 							"<param><value><string>%s</string></value></param>"
 							"</params></methodCall>", (unsigned long) timestamp,
-							logindata->user, logindata->password);
+							logindata->user, xmlencodedPassword);
 					snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
 							"rid=%s&method=authenticateListener", ph->routeId);
+
+					free (xmlencodedPassword);
 					break;
+				}
 			}
 			break;
 		}
 
 		case PIANO_REQUEST_GET_STATIONS:
 			/* get stations, user must be authenticated */
-			assert (ph->user.listenerId != NULL);
-
 			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
 					"<methodCall><methodName>station.getStations</methodName>"
 					"<params><param><value><int>%lu</int></value></param>"
@@ -269,8 +287,7 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=getStations", ph->routeId,
-					ph->user.listenerId);
+					"rid=%s&method=getStations", ph->routeId);
 			break;
 
 		case PIANO_REQUEST_GET_PLAYLIST: {
@@ -305,10 +322,10 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					ph->user.authToken, reqData->station->id,
 					PianoAudioFormatToString (reqData->format));
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=getFragment&arg1=%s&arg2=0"
+					"rid=%s&method=getFragment&arg1=%s&arg2=0"
 					"&arg3=&arg4=&arg5=%s&arg6=0&arg7=0", ph->routeId,
-					ph->user.listenerId, reqData->station->id,
-					PianoAudioFormatToString (reqData->format));
+					reqData->station->id, PianoAudioFormatToString
+					(reqData->format));
 			break;
 		}
 
@@ -347,9 +364,9 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					(reqData->rating == PIANO_RATE_LOVE) ? 1 : 0,
 					reqData->songType);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=addFeedback&arg1=%s&arg2=%s"
+					"rid=%s&method=addFeedback&arg1=%s&arg2=%s"
 					"&arg3=%s&arg4=%u&arg5=%s&arg6=false&arg7=%u",
-					ph->routeId, ph->user.listenerId, reqData->stationId,
+					ph->routeId, reqData->stationId,
 					reqData->musicId,
 					(reqData->userSeed == NULL) ? "" : reqData->userSeed,
 					reqData->testStrategy,
@@ -385,9 +402,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					ph->user.authToken, reqData->station->id,
 					xmlencodedNewName);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=setStationName&arg1=%s&arg2=%s",
-					ph->routeId, ph->user.listenerId, reqData->station->id,
-					urlencodedNewName);
+					"rid=%s&method=setStationName&arg1=%s&arg2=%s",
+					ph->routeId, reqData->station->id, urlencodedNewName);
 
 			free (urlencodedNewName);
 			free (xmlencodedNewName);
@@ -410,8 +426,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, station->id);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=removeStation&arg1=%s", ph->routeId,
-					ph->user.listenerId, station->id);
+					"rid=%s&method=removeStation&arg1=%s", ph->routeId,
+					station->id);
 			break;
 		}
 
@@ -438,8 +454,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, xmlencodedSearchStr);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=search&arg1=%s", ph->routeId,
-					ph->user.listenerId, urlencodedSearchStr);
+					"rid=%s&method=search&arg1=%s", ph->routeId,
+					urlencodedSearchStr);
 
 			free (urlencodedSearchStr);
 			free (xmlencodedSearchStr);
@@ -458,14 +474,18 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
 					"<methodCall><methodName>station.createStation</methodName>"
 					"<params><param><value><int>%lu</int></value></param>"
+					/* auth token */
 					"<param><value><string>%s</string></value></param>"
+					/* music id */
 					"<param><value><string>%s%s</string></value></param>"
+					/* empty */
+					"<param><value><string></string></value></param>"
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, reqData->type, reqData->id);
 
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=createStation&arg1=%s%s", ph->routeId,
-					ph->user.listenerId, reqData->type, reqData->id);
+					"rid=%s&method=createStation&arg1=%s%s&arg2=", ph->routeId,
+					reqData->type, reqData->id);
 			break;
 		}
 
@@ -480,14 +500,17 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
 					"<methodCall><methodName>station.addSeed</methodName><params>"
 					"<param><value><int>%lu</int></value></param>"
+					/* auth token */
 					"<param><value><string>%s</string></value></param>"
+					/* station id */
 					"<param><value><string>%s</string></value></param>"
+					/* music id */
 					"<param><value><string>%s</string></value></param>"
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, reqData->station->id, reqData->musicId);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=addSeed&arg1=%s&arg2=%s", ph->routeId,
-					ph->user.listenerId, reqData->station->id, reqData->musicId);
+					"rid=%s&method=addSeed&arg1=%s&arg2=%s", ph->routeId,
+					reqData->station->id, reqData->musicId);
 			break;
 		}
 
@@ -513,8 +536,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					(song->userSeed == NULL) ? "" : song->userSeed,
 					song->stationId);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=addTiredSong&arg1=%s&arg2=%s&arg3=%s",
-					ph->routeId, ph->user.listenerId,
+					"rid=%s&method=addTiredSong&arg1=%s&arg2=%s&arg3=%s",
+					ph->routeId,
 					(song->musicId == NULL) ? "" : song->musicId,
 					(song->userSeed == NULL) ? "" : song->userSeed,
 					song->stationId);
@@ -559,16 +582,16 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 			}
 			strncat (xmlSendBuf,
 					"</data></array></value></param>"
-					/* content type */
-					"<param><value><string>CUSTOM</string></value></param>"
-					/* genre */
+					/* empty */
+					"<param><value><string></string></value></param>"
+					/* empty */
 					"<param><value><string></string></value></param>"
 					"</params></methodCall>",
 					sizeof (xmlSendBuf) - strlen (xmlSendBuf) - 1);
 
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=setQuickMix&arg1=RANDOM&arg2=%s&arg3=CUSTOM&arg4=",
-					ph->routeId, ph->user.listenerId, urlArgBuf);
+					"rid=%s&method=setQuickMix&arg1=RANDOM&arg2=%s&arg3=&arg4=",
+					ph->routeId, urlArgBuf);
 			break;
 		}
 
@@ -595,8 +618,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, station->id);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=transformShared&arg1=%s", ph->routeId,
-					ph->user.listenerId, station->id);
+					"rid=%s&method=transformShared&arg1=%s", ph->routeId,
+					station->id);
 			break;
 		}
 
@@ -620,8 +643,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					ph->user.authToken, reqData->song->stationId,
 					reqData->song->musicId);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=narrative&arg1=%s&arg2=%s",
-					ph->routeId, ph->user.listenerId, reqData->song->stationId,
+					"rid=%s&method=narrative&arg1=%s&arg2=%s",
+					ph->routeId, reqData->song->stationId,
 					reqData->song->musicId);
 			break;
 		}
@@ -639,15 +662,19 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"<params><param><value><int>%lu</int></value></param>"
 					/* auth token */
 					"<param><value><string>%s</string></value></param>"
+					/* station id */
+					"<param><value><string>%s</string></value></param>"
 					/* seed music id */
 					"<param><value><string>%s</string></value></param>"
 					/* max */
 					"<param><value><int>%u</int></value></param>"
 					"</params></methodCall>", (unsigned long) timestamp,
-					ph->user.authToken, reqData->musicId, reqData->max);
+					ph->user.authToken, reqData->station->id, reqData->musicId,
+					reqData->max);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=getSeedSuggestions&arg1=%s&arg2=%u",
-					ph->routeId, ph->user.listenerId, reqData->musicId, reqData->max);
+					"rid=%s&method=getSeedSuggestions&arg1=%s&arg2=%s&arg3=%u",
+					ph->routeId, reqData->station->id, reqData->musicId,
+					reqData->max);
 			break;
 		}
 
@@ -669,9 +696,8 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, song->stationId, song->musicId);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=createBookmark&arg1=%s&arg2=%s",
-					ph->routeId, ph->user.listenerId, song->stationId,
-					song->musicId);
+					"rid=%s&method=createBookmark&arg1=%s&arg2=%s",
+					ph->routeId, song->stationId, song->musicId);
 			break;
 		}
 
@@ -691,8 +717,80 @@ PianoReturn_t PianoRequest (PianoHandle_t *ph, PianoRequest_t *req,
 					"</params></methodCall>", (unsigned long) timestamp,
 					ph->user.authToken, song->artistMusicId);
 			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
-					"rid=%s&lid=%s&method=createArtistBookmark&arg1=%s",
-					ph->routeId, ph->user.listenerId, song->artistMusicId);
+					"rid=%s&method=createArtistBookmark&arg1=%s",
+					ph->routeId, song->artistMusicId);
+			break;
+		}
+
+		case PIANO_REQUEST_GET_STATION_INFO: {
+			/* get station information (seeds and feedback) */
+			PianoRequestDataGetStationInfo_t *reqData = req->data;
+
+			assert (reqData != NULL);
+			assert (reqData->station != NULL);
+
+			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
+					"<methodCall><methodName>station.getStation</methodName>"
+					"<params><param><value><int>%lu</int></value></param>"
+					/* auth token */
+					"<param><value><string>%s</string></value></param>"
+					/* station id */
+					"<param><value><string>%s</string></value></param>"
+					"</params></methodCall>", (unsigned long) timestamp,
+					ph->user.authToken, reqData->station->id);
+			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
+					"rid=%s&method=getStation&arg1=%s",
+					ph->routeId, reqData->station->id);
+			break;
+		}
+
+		case PIANO_REQUEST_DELETE_FEEDBACK: {
+			PianoSong_t *song = req->data;
+
+			assert (song != NULL);
+
+			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
+					"<methodCall><methodName>station.deleteFeedback</methodName>"
+					"<params><param><value><int>%lu</int></value></param>"
+					/* auth token */
+					"<param><value><string>%s</string></value></param>"
+					/* feedback id */
+					"<param><value><string>%s</string></value></param>"
+					"</params></methodCall>", (unsigned long) timestamp,
+					ph->user.authToken, song->feedbackId);
+			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
+					"rid=%s&method=deleteFeedback&arg1=%s",
+					ph->routeId, song->feedbackId);
+			break;
+		}
+
+		case PIANO_REQUEST_DELETE_SEED: {
+			PianoRequestDataDeleteSeed_t *reqData = req->data;
+			char *seedId = NULL;
+
+			assert (reqData != NULL);
+			assert (reqData->song != NULL || reqData->artist != NULL);
+
+			if (reqData->song != NULL) {
+				seedId = reqData->song->seedId;
+			} else if (reqData->artist != NULL) {
+				seedId = reqData->artist->seedId;
+			}
+
+			assert (seedId != NULL);
+
+			snprintf (xmlSendBuf, sizeof (xmlSendBuf), "<?xml version=\"1.0\"?>"
+					"<methodCall><methodName>station.deleteSeed</methodName>"
+					"<params><param><value><int>%lu</int></value></param>"
+					/* auth token */
+					"<param><value><string>%s</string></value></param>"
+					/* seed id */
+					"<param><value><string>%s</string></value></param>"
+					"</params></methodCall>", (unsigned long) timestamp,
+					ph->user.authToken, seedId);
+			snprintf (req->urlPath, sizeof (req->urlPath), PIANO_RPC_PATH
+					"rid=%s&method=deleteSeed&arg1=%s",
+					ph->routeId, seedId);
 			break;
 		}
 
@@ -825,7 +923,7 @@ PianoReturn_t PianoResponse (PianoHandle_t *ph, PianoRequest_t *req) {
 				case 1:
 					/* information exists when reauthenticating, destroy to
 					 * avoid memleak */
-					if (ph->user.listenerId != NULL) {
+					if (ph->user.authToken != NULL) {
 						PianoDestroyUserInfo (&ph->user);
 					}
 					ret = PianoXmlParseUserinfo (ph, req->responseData);
@@ -967,6 +1065,7 @@ PianoReturn_t PianoResponse (PianoHandle_t *ph, PianoRequest_t *req) {
 		case PIANO_REQUEST_SET_QUICKMIX:
 		case PIANO_REQUEST_BOOKMARK_SONG:
 		case PIANO_REQUEST_BOOKMARK_ARTIST:
+		case PIANO_REQUEST_DELETE_FEEDBACK:
 			assert (req->responseData != NULL);
 
 			ret = PianoXmlParseSimple (req->responseData);
@@ -1017,6 +1116,25 @@ PianoReturn_t PianoResponse (PianoHandle_t *ph, PianoRequest_t *req) {
 					&reqData->searchResult);
 			break;
 		}
+
+		case PIANO_REQUEST_GET_STATION_INFO: {
+			/* get station information (seeds and feedback) */
+			PianoRequestDataGetStationInfo_t *reqData = req->data;
+
+			assert (req->responseData != NULL);
+			assert (reqData != NULL);
+
+			ret = PianoXmlParseGetStationInfo (req->responseData,
+					&reqData->info);
+			break;
+		}
+
+		case PIANO_REQUEST_DELETE_SEED: {
+			assert (req->responseData != NULL);
+
+			/* dummy function, checks for errors only */
+			ret = PianoXmlParseTranformStation (req->responseData);
+		}
 	}
 
 	return ret;
@@ -1106,6 +1224,10 @@ const char *PianoErrorToStr (PianoReturn_t ret) {
 			return "Quickmix not playable.";
 			break;
 
+		case PIANO_RET_REMOVING_TOO_MANY_SEEDS:
+			return "Last seed cannot be removed.";
+			break;
+
 		default:
 			return "No error message available.";
 			break;
