@@ -35,11 +35,15 @@ THE SOFTWARE.
 #endif
 
 #include <ao/ao.h>
+/* required for freebsd */
+#include <sys/types.h>
 #include <pthread.h>
 
 #include <piano.h>
 #include <waitress.h>
 
+#include "settings.h"
+
 #define BAR_PLAYER_MS_TO_S_FACTOR 1000
 
 struct audioPlayer {
@@ -97,6 +101,8 @@ struct audioPlayer {
 
 	char doQuit;
 	pthread_mutex_t pauseMutex;
+
+	const BarSettings_t *settings;
 };
 
 enum {PLAYER_RET_OK = 0, PLAYER_RET_ERR = 1};
