@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2008-2010
+Copyright (c) 2008-2011
 	Lars-Dominik Braun <lars@6xq.net>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -23,8 +23,11 @@ THE SOFTWARE.
 
 /* everything that interacts with the user */
 
+#ifndef __FreeBSD__
 #define _POSIX_C_SOURCE 1 /* fileno() */
 #define _BSD_SOURCE /* strdup() */
+#define _DARWIN_C_SOURCE /* strdup() on OS X */
+#endif
 
 #include <stdio.h>
 #include <stdarg.h>
@@ -34,6 +37,7 @@ THE SOFTWARE.
 #include <errno.h>
 #include <strings.h>
 #include <assert.h>
+#include <ctype.h> /* tolower() */
 
 /* waitpid () */
 #include <sys/types.h>
@@ -44,60 +48,87 @@ THE SOFTWARE.
 
 typedef int (*BarSortFunc_t) (const void *, const void *);
 
+/*	is string a number?
+ */
+static bool isnumeric (const char *s) {
+	if (*s == '\0') {
+		return false;
+	}
+	while (*s != '\0') {
+		if (!isdigit ((unsigned char) *s)) {
+			return false;
+		}
+		++s;
+	}
+	return true;
+}
+
+/*	find needle in haystack, ignoring case, and return first position
+ */
+static const char *BarStrCaseStr (const char *haystack, const char *needle) {
+	const char *needlePos = needle;
+
+	assert (haystack != NULL);
+	assert (needle != NULL);
+
+	if (*needle == '\0') {
+		return haystack;
+	}
+
+	while (*haystack != '\0') {
+		if (tolower ((unsigned char) *haystack) == tolower ((unsigned char) *needlePos)) {
+			++needlePos;
+		} else {
+			needlePos = needle;
+		}
+		++haystack;
+		if (*needlePos == '\0') {
+			return haystack - strlen (needle);
+		}
+	}
+
+	return NULL;
+}
+
 /*	output message and flush stdout
  *	@param message
  */
-inline void BarUiMsg (uiMsg_t type, const char *format, ...) {
-	#define ANSI_CLEAR_LINE "\033[2K"
+void BarUiMsg (const BarSettings_t *settings, const BarUiMsg_t type,
+		const char *format, ...) {
 	va_list fmtargs;
 
+	assert (settings != NULL);
+	assert (type < MSG_COUNT);
+	assert (format != NULL);
+
 	switch (type) {
 		case MSG_INFO:
-			printf (ANSI_CLEAR_LINE "(i) ");
-			break;
-
 		case MSG_PLAYING:
-			printf (ANSI_CLEAR_LINE "|>  ");
-			break;
-
 		case MSG_TIME:
-			printf (ANSI_CLEAR_LINE "#   ");
-			break;
-		
 		case MSG_ERR:
-			printf (ANSI_CLEAR_LINE "/!\\ ");
-			break;
-
 		case MSG_QUESTION:
-			printf (ANSI_CLEAR_LINE "[?] ");
+		case MSG_LIST:
+			/* print ANSI clear line */
+			fputs ("\033[2K", stdout);
 			break;
 
-		case MSG_LIST:
-			printf (ANSI_CLEAR_LINE "\t");
-			break;
-	
 		default:
 			break;
 	}
+
+	if (settings->msgFormat[type].prefix != NULL) {
+		fputs (settings->msgFormat[type].prefix, stdout);
+	}
+
 	va_start (fmtargs, format);
 	vprintf (format, fmtargs);
 	va_end (fmtargs);
 
-	fflush (stdout);
-
-	#undef ANSI_CLEAR_LINE
-}
-
-/*	prints human readable status message based on return value
- *	@param piano return value
- */
-inline PianoReturn_t BarUiPrintPianoStatus (PianoReturn_t ret) {
-	if (ret != PIANO_RET_OK) {
-		BarUiMsg (MSG_NONE, "Error: %s\n", PianoErrorToStr (ret));
-	} else {
-		BarUiMsg (MSG_NONE, "Ok.\n");
+	if (settings->msgFormat[type].postfix != NULL) {
+		fputs (settings->msgFormat[type].postfix, stdout);
 	}
-	return ret;
+
+	fflush (stdout);
 }
 
 /*	fetch http resource (post request)
@@ -109,17 +140,18 @@ static WaitressReturn_t BarPianoHttpRequest (WaitressHandle_t *waith,
 	waith->extraHeaders = "Content-Type: text/xml\r\n";
 	waith->postData = req->postData;
 	waith->method = WAITRESS_METHOD_POST;
-	strncpy (waith->path, req->urlPath, sizeof (waith->path)-1);
+	waith->url.path = req->urlPath;
 
 	return WaitressFetchBuf (waith, &req->responseData);
 }
 
 /*	piano wrapper: prepare/execute http request and pass result back to
  *	libpiano (updates data structures)
- *	@param piano handle
+ *	@param app handle
  *	@param request type
- *	@param waitress handle
- *	@param data pointer (used as request data)
+ *	@param request data
+ *	@param stores piano return code
+ *	@param stores waitress return code
  *	@return 1 on success, 0 otherwise
  */
 int BarUiPianoCall (BarApp_t * const app, PianoRequestType_t type,
@@ -134,14 +166,14 @@ int BarUiPianoCall (BarApp_t * const app, PianoRequestType_t type,
 
 		*pRet = PianoRequest (&app->ph, &req, type);
 		if (*pRet != PIANO_RET_OK) {
-			BarUiMsg (MSG_NONE, "Error: %s\n", PianoErrorToStr (*pRet));
+			BarUiMsg (&app->settings, MSG_NONE, "Error: %s\n", PianoErrorToStr (*pRet));
 			PianoDestroyRequest (&req);
 			return 0;
 		}
 
 		*wRet = BarPianoHttpRequest (&app->waith, &req);
 		if (*wRet != WAITRESS_RET_OK) {
-			BarUiMsg (MSG_NONE, "Network error: %s\n", WaitressErrorToStr (*wRet));
+			BarUiMsg (&app->settings, MSG_NONE, "Network error: %s\n", WaitressErrorToStr (*wRet));
 			if (req.responseData != NULL) {
 				free (req.responseData);
 			}
@@ -162,7 +194,7 @@ int BarUiPianoCall (BarApp_t * const app, PianoRequestType_t type,
 				reqData.password = app->settings.password;
 				reqData.step = 0;
 
-				BarUiMsg (MSG_NONE, "Reauthentication required... ");
+				BarUiMsg (&app->settings, MSG_NONE, "Reauthentication required... ");
 				if (!BarUiPianoCall (app, PIANO_REQUEST_LOGIN, &reqData, &authpRet,
 						&authwRet)) {
 					*pRet = authpRet;
@@ -175,17 +207,17 @@ int BarUiPianoCall (BarApp_t * const app, PianoRequestType_t type,
 				} else {
 					/* try again */
 					*pRet = PIANO_RET_CONTINUE_REQUEST;
-					BarUiMsg (MSG_INFO, "Trying again... ");
+					BarUiMsg (&app->settings, MSG_INFO, "Trying again... ");
 				}
 			} else if (*pRet != PIANO_RET_OK) {
-				BarUiMsg (MSG_NONE, "Error: %s\n", PianoErrorToStr (*pRet));
+				BarUiMsg (&app->settings, MSG_NONE, "Error: %s\n", PianoErrorToStr (*pRet));
 				if (req.responseData != NULL) {
 					free (req.responseData);
 				}
 				PianoDestroyRequest (&req);
 				return 0;
 			} else {
-				BarUiMsg (MSG_NONE, "Ok.\n");
+				BarUiMsg (&app->settings, MSG_NONE, "Ok.\n");
 			}
 		}
 		/* we can destroy the request at this point, even when this call needs
@@ -258,7 +290,7 @@ static int BarStationCmpQuickmix10NameZA (const void *a, const void *b) {
  *	@param stations
  *	@return NULL-terminated array with sorted stations
  */
-PianoStation_t **BarSortedStations (PianoStation_t *unsortedStations,
+static PianoStation_t **BarSortedStations (PianoStation_t *unsortedStations,
 		size_t *retStationCount, BarStationSorting_t order) {
 	static const BarSortFunc_t orderMapping[] = {BarStationNameAZCmp,
 			BarStationNameZACmp,
@@ -296,40 +328,61 @@ PianoStation_t **BarSortedStations (PianoStation_t *unsortedStations,
 }
 
 /*	let user pick one station
- *	@param piano handle
+ *	@param app handle
+ *	@param prompt string
+ *	@param called if input was not a number
  *	@return pointer to selected station or NULL
  */
-PianoStation_t *BarUiSelectStation (PianoHandle_t *ph, const char *prompt,
-		BarStationSorting_t order, FILE *curFd) {
+PianoStation_t *BarUiSelectStation (BarApp_t *app, const char *prompt,
+		BarUiSelectStationCallback_t callback) {
 	PianoStation_t **sortedStations = NULL, *retStation = NULL;
 	size_t stationCount, i;
-	int input;
+	char buf[100];
 
-	if (ph->stations == NULL) {
-		BarUiMsg (MSG_ERR, "No station available.\n");
+	if (app->ph.stations == NULL) {
+		BarUiMsg (&app->settings, MSG_ERR, "No station available.\n");
 		return NULL;
 	}
 
+	memset (buf, 0, sizeof (buf));
+
 	/* sort and print stations */
-	sortedStations = BarSortedStations (ph->stations, &stationCount, order);
-	for (i = 0; i < stationCount; i++) {
-		const PianoStation_t *currStation = sortedStations[i];
-		BarUiMsg (MSG_LIST, "%2i) %c%c%c %s\n", i,
-				currStation->useQuickMix ? 'q' : ' ',
-				currStation->isQuickMix ? 'Q' : ' ',
-				!currStation->isCreator ? 'S' : ' ',
-				currStation->name);
-	}
+	sortedStations = BarSortedStations (app->ph.stations, &stationCount,
+			app->settings.sortOrder);
+
+	do {
+		for (i = 0; i < stationCount; i++) {
+			const PianoStation_t *currStation = sortedStations[i];
+			/* filter stations */
+			if (BarStrCaseStr (currStation->name, buf) != NULL) {
+				BarUiMsg (&app->settings, MSG_LIST, "%2i) %c%c%c %s\n", i,
+						currStation->useQuickMix ? 'q' : ' ',
+						currStation->isQuickMix ? 'Q' : ' ',
+						!currStation->isCreator ? 'S' : ' ',
+						currStation->name);
+			}
+		}
+
+		BarUiMsg (&app->settings, MSG_QUESTION, prompt);
+		if (BarReadlineStr (buf, sizeof (buf), &app->input,
+				BAR_RL_DEFAULT) == 0) {
+			free (sortedStations);
+			return NULL;
+		}
+
+		if (isnumeric (buf)) {
+			unsigned long selected = strtoul (buf, NULL, 0);
+			if (selected < stationCount) {
+				retStation = sortedStations[selected];
+			}
+		}
+
+		/* hand over buffer to external function if it was not a station number */
+		if (retStation == NULL && callback != NULL) {
+			callback (app, buf);
+		}
+	} while (retStation == NULL);
 
-	BarUiMsg (MSG_QUESTION, prompt);
-	/* FIXME: using a _signed_ int is ugly */
-	if (BarReadlineInt (&input, curFd) == 0) {
-		free (sortedStations);
-		return NULL;
-	}
-	if (input < stationCount) {
-		retStation = sortedStations[input];
-	}
 	free (sortedStations);
 	return retStation;
 }
@@ -337,81 +390,109 @@ PianoStation_t *BarUiSelectStation (PianoHandle_t *ph, const char *prompt,
 /*	let user pick one song
  *	@param pianobar settings
  *	@param song list
- *	@param current fd
+ *	@param input fds
  *	@return pointer to selected item in song list or NULL
  */
 PianoSong_t *BarUiSelectSong (const BarSettings_t *settings,
-		PianoSong_t *startSong, FILE *curFd) {
+		PianoSong_t *startSong, BarReadlineFds_t *input) {
 	PianoSong_t *tmpSong = NULL;
-	int i = 0;
+	char buf[100];
 
-	i = BarUiListSongs (settings, startSong);
+	memset (buf, 0, sizeof (buf));
 
-	BarUiMsg (MSG_QUESTION, "Select song: ");
-	if (BarReadlineInt (&i, curFd) == 0) {
-		return NULL;
-	}
+	do {
+		BarUiListSongs (settings, startSong, buf);
 
-	tmpSong = startSong;
-	while (tmpSong != NULL && i > 0) {
-		tmpSong = tmpSong->next;
-		i--;
-	}
+		BarUiMsg (settings, MSG_QUESTION, "Select song: ");
+		if (BarReadlineStr (buf, sizeof (buf), input, BAR_RL_DEFAULT) == 0) {
+			return NULL;
+		}
+
+		if (isnumeric (buf)) {
+			unsigned long i = strtoul (buf, NULL, 0);
+			tmpSong = startSong;
+			while (tmpSong != NULL && i > 0) {
+				tmpSong = tmpSong->next;
+				i--;
+			}
+		}
+	} while (tmpSong == NULL);
 
 	return tmpSong;
 }
 
 /*	let user pick one artist
+ *	@param app handle
  *	@param artists (linked list)
  *	@return pointer to selected artist or NULL on abort
  */
-PianoArtist_t *BarUiSelectArtist (PianoArtist_t *startArtist, FILE *curFd) {
+PianoArtist_t *BarUiSelectArtist (BarApp_t *app, PianoArtist_t *startArtist) {
 	PianoArtist_t *tmpArtist = NULL;
-	int i = 0;
+	char buf[100];
+	unsigned long i;
+
+	memset (buf, 0, sizeof (buf));
+
+	do {
+		/* print all artists */
+		i = 0;
+		tmpArtist = startArtist;
+		while (tmpArtist != NULL) {
+			if (BarStrCaseStr (tmpArtist->name, buf) != NULL) {
+				BarUiMsg (&app->settings, MSG_LIST, "%2u) %s\n", i, tmpArtist->name);
+			}
+			i++;
+			tmpArtist = tmpArtist->next;
+		}
+
+		BarUiMsg (&app->settings, MSG_QUESTION, "Select artist: ");
+		if (BarReadlineStr (buf, sizeof (buf), &app->input,
+				BAR_RL_DEFAULT) == 0) {
+			return NULL;
+		}
+
+		if (isnumeric (buf)) {
+			i = strtoul (buf, NULL, 0);
+			tmpArtist = startArtist;
+			while (tmpArtist != NULL && i > 0) {
+				tmpArtist = tmpArtist->next;
+				i--;
+			}
+		}
+	} while (tmpArtist == NULL);
 
-	/* print all artists */
-	tmpArtist = startArtist;
-	while (tmpArtist != NULL) {
-		BarUiMsg (MSG_LIST, "%2u) %s\n", i, tmpArtist->name);
-		i++;
-		tmpArtist = tmpArtist->next;
-	}
-	BarUiMsg (MSG_QUESTION, "Select artist: ");
-	if (BarReadlineInt (&i, curFd) == 0) {
-		return NULL;
-	}
-	tmpArtist = startArtist;
-	while (tmpArtist != NULL && i > 0) {
-		tmpArtist = tmpArtist->next;
-		i--;
-	}
 	return tmpArtist;
 }
 
 /*	search music: query, search request, return music id
- *	@param piano handle
- *	@param read data from fd
- *	@param allow seed suggestions if != NULL
+ *	@param app handle
+ *	@param seed suggestion station
+ *	@param seed suggestion musicid
+ *	@param prompt string
  *	@return musicId or NULL on abort/error
  */
-char *BarUiSelectMusicId (BarApp_t *app, FILE *curFd, char *similarToId) {
+char *BarUiSelectMusicId (BarApp_t *app, PianoStation_t *station,
+		PianoSong_t *similarSong, const char *msg) {
 	char *musicId = NULL;
 	char lineBuf[100], selectBuf[2];
 	PianoSearchResult_t searchResult;
 	PianoArtist_t *tmpArtist;
 	PianoSong_t *tmpSong;
 
-	BarUiMsg (MSG_QUESTION, "Search for artist/title: ");
-	if (BarReadlineStr (lineBuf, sizeof (lineBuf), 0, curFd) > 0) {
-		if (strcmp ("?", lineBuf) == 0 && similarToId != NULL) {
+	BarUiMsg (&app->settings, MSG_QUESTION, msg);
+	if (BarReadlineStr (lineBuf, sizeof (lineBuf), &app->input,
+			BAR_RL_DEFAULT) > 0) {
+		if (strcmp ("?", lineBuf) == 0 && station != NULL &&
+				similarSong != NULL) {
 			PianoReturn_t pRet;
 			WaitressReturn_t wRet;
 			PianoRequestDataGetSeedSuggestions_t reqData;
 
-			reqData.musicId = similarToId;
+			reqData.station = station;
+			reqData.musicId = similarSong->musicId;
 			reqData.max = 20;
 
-			BarUiMsg (MSG_INFO, "Receiving suggestions... ");
+			BarUiMsg (&app->settings, MSG_INFO, "Receiving suggestions... ");
 			if (!BarUiPianoCall (app, PIANO_REQUEST_GET_SEED_SUGGESTIONS,
 					&reqData, &pRet, &wRet)) {
 				return NULL;
@@ -424,27 +505,28 @@ char *BarUiSelectMusicId (BarApp_t *app, FILE *curFd, char *similarToId) {
 
 			reqData.searchStr = lineBuf;
 
-			BarUiMsg (MSG_INFO, "Searching... ");
+			BarUiMsg (&app->settings, MSG_INFO, "Searching... ");
 			if (!BarUiPianoCall (app, PIANO_REQUEST_SEARCH, &reqData, &pRet,
 					&wRet)) {
 				return NULL;
 			}
 			memcpy (&searchResult, &reqData.searchResult, sizeof (searchResult));
 		}
-		BarUiMsg (MSG_NONE, "\r");
+		BarUiMsg (&app->settings, MSG_NONE, "\r");
 		if (searchResult.songs != NULL &&
 				searchResult.artists != NULL) {
 			/* songs and artists found */
-			BarUiMsg (MSG_QUESTION, "Is this an [a]rtist or [t]rack name? ");
-			BarReadline (selectBuf, sizeof (selectBuf), "at", 1, 0, curFd);
+			BarUiMsg (&app->settings, MSG_QUESTION, "Is this an [a]rtist or [t]rack name? ");
+			BarReadline (selectBuf, sizeof (selectBuf), "at", &app->input,
+					BAR_RL_FULLRETURN, -1);
 			if (*selectBuf == 'a') {
-				tmpArtist = BarUiSelectArtist (searchResult.artists, curFd);
+				tmpArtist = BarUiSelectArtist (app, searchResult.artists);
 				if (tmpArtist != NULL) {
 					musicId = strdup (tmpArtist->musicId);
 				}
 			} else if (*selectBuf == 't') {
 				tmpSong = BarUiSelectSong (&app->settings, searchResult.songs,
-						curFd);
+						&app->input);
 				if (tmpSong != NULL) {
 					musicId = strdup (tmpSong->musicId);
 				}
@@ -452,18 +534,18 @@ char *BarUiSelectMusicId (BarApp_t *app, FILE *curFd, char *similarToId) {
 		} else if (searchResult.songs != NULL) {
 			/* songs found */
 			tmpSong = BarUiSelectSong (&app->settings, searchResult.songs,
-					curFd);
+					&app->input);
 			if (tmpSong != NULL) {
 				musicId = strdup (tmpSong->musicId);
 			}
 		} else if (searchResult.artists != NULL) {
 			/* artists found */
-			tmpArtist = BarUiSelectArtist (searchResult.artists, curFd);
+			tmpArtist = BarUiSelectArtist (app, searchResult.artists);
 			if (tmpArtist != NULL) {
 				musicId = strdup (tmpArtist->musicId);
 			}
 		} else {
-			BarUiMsg (MSG_INFO, "Nothing found...\n");
+			BarUiMsg (&app->settings, MSG_INFO, "Nothing found...\n");
 		}
 		PianoDestroySearchResult (&searchResult);
 	}
@@ -472,9 +554,9 @@ char *BarUiSelectMusicId (BarApp_t *app, FILE *curFd, char *similarToId) {
 }
 
 /*	browse genre stations and create shared station
- *	@param piano handle
+ *	@param app handle
  */
-void BarStationFromGenre (BarApp_t *app, FILE *curFd) {
+void BarStationFromGenre (BarApp_t *app) {
 	PianoReturn_t pRet;
 	WaitressReturn_t wRet;
 	PianoGenreCategory_t *curCat;
@@ -487,7 +569,7 @@ void BarStationFromGenre (BarApp_t *app, FILE *curFd) {
 		PianoReturn_t pRet;
 		WaitressReturn_t wRet;
 
-		BarUiMsg (MSG_INFO, "Receiving genre stations... ");
+		BarUiMsg (&app->settings, MSG_INFO, "Receiving genre stations... ");
 		if (!BarUiPianoCall (app, PIANO_REQUEST_GET_GENRE_STATIONS, NULL,
 				&pRet, &wRet)) {
 			return;
@@ -498,13 +580,13 @@ void BarStationFromGenre (BarApp_t *app, FILE *curFd) {
 	curCat = app->ph.genreStations;
 	i = 0;
 	while (curCat != NULL) {
-		BarUiMsg (MSG_LIST, "%2i) %s\n", i, curCat->name);
+		BarUiMsg (&app->settings, MSG_LIST, "%2i) %s\n", i, curCat->name);
 		i++;
 		curCat = curCat->next;
 	}
 	/* select category or exit */
-	BarUiMsg (MSG_QUESTION, "Select category: ");
-	if (BarReadlineInt (&i, curFd) == 0) {
+	BarUiMsg (&app->settings, MSG_QUESTION, "Select category: ");
+	if (BarReadlineInt (&i, &app->input) == 0) {
 		return;
 	}
 	curCat = app->ph.genreStations;
@@ -517,12 +599,12 @@ void BarStationFromGenre (BarApp_t *app, FILE *curFd) {
 	curGenre = curCat->genres;
 	i = 0;
 	while (curGenre != NULL) {
-		BarUiMsg (MSG_LIST, "%2i) %s\n", i, curGenre->name);
+		BarUiMsg (&app->settings, MSG_LIST, "%2i) %s\n", i, curGenre->name);
 		i++;
 		curGenre = curGenre->next;
 	}
-	BarUiMsg (MSG_QUESTION, "Select genre: ");
-	if (BarReadlineInt (&i, curFd) == 0) {
+	BarUiMsg (&app->settings, MSG_QUESTION, "Select genre: ");
+	if (BarReadlineInt (&i, &app->input) == 0) {
 		return;
 	}
 	curGenre = curCat->genres;
@@ -531,17 +613,98 @@ void BarStationFromGenre (BarApp_t *app, FILE *curFd) {
 		i--;
 	}
 	/* create station */
-	BarUiMsg (MSG_INFO, "Adding shared station \"%s\"... ", curGenre->name);
+	BarUiMsg (&app->settings, MSG_INFO, "Adding shared station \"%s\"... ", curGenre->name);
 	reqData.id = curGenre->musicId;
 	reqData.type = "mi";
 	BarUiPianoCall (app, PIANO_REQUEST_CREATE_STATION, &reqData, &pRet, &wRet);
 }
 
-/*	Print station infos (including station id)
+/*	replaces format characters (%x) in format string with custom strings
+ *	@param destination buffer
+ *	@param dest buffer size
+ *	@param format string
+ *	@param format characters
+ *	@param replacement for each given format character
+ */
+void BarUiCustomFormat (char *dest, size_t destSize, const char *format,
+		const char *formatChars, const char **formatVals) {
+	bool haveFormatChar = false;
+
+	while (*format != '\0' && destSize > 1) {
+		if (*format == '%' && !haveFormatChar) {
+			haveFormatChar = true;
+		} else if (haveFormatChar) {
+			const char *testChar = formatChars;
+			const char *val = NULL;
+
+			/* search for format character */
+			while (*testChar != '\0') {
+				if (*testChar == *format) {
+					val = formatVals[(testChar-formatChars)/sizeof (*testChar)];
+					break;
+				}
+				++testChar;
+			}
+
+			if (val != NULL) {
+				/* concat */
+				while (*val != '\0' && destSize > 1) {
+					*dest = *val;
+					++val;
+					++dest;
+					--destSize;
+				}
+			} else {
+				/* invalid format character */
+				*dest = '%';
+				++dest;
+				--destSize;
+				if (destSize > 1) {
+					*dest = *format;
+					++dest;
+					--destSize;
+				}
+			}
+
+			haveFormatChar = false;
+		} else {
+			/* copy */
+			*dest = *format;
+			++dest;
+			--destSize;
+		}
+		++format;
+	}
+	*dest = '\0';
+}
+
+/*	append \n to string
+ */
+static void BarUiAppendNewline (char *s, size_t maxlen) {
+	size_t len;
+
+	/* append \n */
+	if ((len = strlen (s)) == maxlen-1) {
+		s[maxlen-2] = '\n';
+	} else {
+		s[len] = '\n';
+		s[len+1] = '\0';
+	}
+}
+
+/*	Print customizeable station infos
+ *	@param pianobar settings
  *	@param the station
  */
-inline void BarUiPrintStation (PianoStation_t *station) {
-	BarUiMsg (MSG_PLAYING, "Station \"%s\" (%s)\n", station->name, station->id);
+inline void BarUiPrintStation (const BarSettings_t *settings,
+		PianoStation_t *station) {
+	char outstr[512];
+	const char *vals[] = {station->name, station->id};
+
+	BarUiCustomFormat (outstr, sizeof (outstr), settings->npStationFormat,
+			"ni", vals);
+	BarUiAppendNewline (outstr, sizeof (outstr));
+	BarUiMsg (settings, MSG_PLAYING, outstr);
 }
 
 /*	Print song infos (artist, title, album, loved)
@@ -551,30 +714,61 @@ inline void BarUiPrintStation (PianoStation_t *station) {
  */
 inline void BarUiPrintSong (const BarSettings_t *settings,
 		const PianoSong_t *song, const PianoStation_t *station) {
-	BarUiMsg (MSG_PLAYING, "\"%s\" by \"%s\" on \"%s\"%s%s%s%s\n",
-			song->title, song->artist, song->album,
-			(song->rating == PIANO_RATE_LOVE) ? " " : "",
+	char outstr[512];
+	const char *vals[] = {song->title, song->artist, song->album,
 			(song->rating == PIANO_RATE_LOVE) ? settings->loveIcon : "",
-			station != NULL ? " @ " : "",
-			station != NULL ? station->name : "");
+			station != NULL ? settings->atIcon : "",
+			station != NULL ? station->name : "",
+			song->detailUrl};
+
+	BarUiCustomFormat (outstr, sizeof (outstr), settings->npSongFormat,
+			"talr@su", vals);
+	BarUiAppendNewline (outstr, sizeof (outstr));
+	BarUiMsg (settings, MSG_PLAYING, outstr);
+}
+
+/*	Incremets two digit ASCII counter
+ */
+static void BarUiIncDigits (char digits[3]) {
+	++digits[1];
+	if (digits[1] == ':') {
+		digits[1] = '0';
+		digits[0] |= 0x30;
+		++digits[0];
+		if (digits[0] == ':') {
+			digits[0] = ' ';
+		}
+	}
 }
 
 /*	Print list of songs
  *	@param pianobar settings
  *	@param linked list of songs
+ *	@param artist/song filter string
  *	@return # of songs
  */
 size_t BarUiListSongs (const BarSettings_t *settings,
-		const PianoSong_t *song) {
+		const PianoSong_t *song, const char *filter) {
 	size_t i = 0;
+	char digits[3] = " 0";
 
 	while (song != NULL) {
-		BarUiMsg (MSG_LIST, "%2lu) %s - %s %s%s\n", i, song->artist,
-				song->title,
-				(song->rating == PIANO_RATE_LOVE) ? settings->loveIcon : "",
-				(song->rating == PIANO_RATE_BAN) ? settings->banIcon : "");
-		song = song->next;
+		if (filter == NULL ||
+				(filter != NULL && (BarStrCaseStr (song->artist, filter) != NULL ||
+				BarStrCaseStr (song->title, filter) != NULL))) {
+			char outstr[512];
+			const char *vals[] = {digits, song->artist, song->title,
+					(song->rating == PIANO_RATE_LOVE) ? settings->loveIcon :
+					((song->rating == PIANO_RATE_BAN) ? settings->banIcon : "")};
+
+			BarUiCustomFormat (outstr, sizeof (outstr), settings->listSongFormat,
+					"iatr", vals);
+			BarUiAppendNewline (outstr, sizeof (outstr));
+			BarUiMsg (settings, MSG_LIST, outstr);
+		}
 		i++;
+		BarUiIncDigits (digits);
+		song = song->next;
 	}
 
 	return i;
@@ -593,7 +787,6 @@ void BarUiStartEventCmd (const BarSettings_t *settings, const char *type,
 		const struct audioPlayer *player, PianoStation_t *stations,
                 PianoReturn_t pRet, WaitressReturn_t wRet) {
 	pid_t chld;
-	char pipeBuf[1024];
 	int pipeFd[2];
 
 	if (settings->eventCmd == NULL) {
@@ -601,37 +794,8 @@ void BarUiStartEventCmd (const BarSettings_t *settings, const char *type,
 		return;
 	}
 
-	/* prepare stdin content */
-	memset (pipeBuf, 0, sizeof (pipeBuf));
-	snprintf (pipeBuf, sizeof (pipeBuf),
-			"artist=%s\n"
-			"title=%s\n"
-			"album=%s\n"
-			"coverArt=%s\n"
-			"stationName=%s\n"
-			"pRet=%i\n"
-			"pRetStr=%s\n"
-			"wRet=%i\n"
-			"wRetStr=%s\n"
-			"songDuration=%lu\n"
-			"songPlayed=%lu\n"
-			"rating=%i\n",
-			curSong == NULL ? "" : curSong->artist,
-			curSong == NULL ? "" : curSong->title,
-			curSong == NULL ? "" : curSong->album,
-			curSong == NULL ? "" : curSong->coverArt,
-			curStation == NULL ? "" : curStation->name,
-			pRet,
-			PianoErrorToStr (pRet),
-			wRet,
-			WaitressErrorToStr (wRet),
-			player->songDuration,
-			player->songPlayed,
-			curSong == NULL ? PIANO_RATE_NONE : curSong->rating
-			);
-
 	if (pipe (pipeFd) == -1) {
-		BarUiMsg (MSG_ERR, "Cannot create eventcmd pipe. (%s)\n", strerror (errno));
+		BarUiMsg (settings, MSG_ERR, "Cannot create eventcmd pipe. (%s)\n", strerror (errno));
 		return;
 	}
 
@@ -641,16 +805,55 @@ void BarUiStartEventCmd (const BarSettings_t *settings, const char *type,
 		close (pipeFd[1]);
 		dup2 (pipeFd[0], fileno (stdin));
 		execl (settings->eventCmd, settings->eventCmd, type, (char *) NULL);
-		BarUiMsg (MSG_ERR, "Cannot start eventcmd. (%s)\n", strerror (errno));
+		BarUiMsg (settings, MSG_ERR, "Cannot start eventcmd. (%s)\n", strerror (errno));
 		close (pipeFd[0]);
 		exit (1);
 	} else if (chld == -1) {
-		BarUiMsg (MSG_ERR, "Cannot fork eventcmd. (%s)\n", strerror (errno));
+		BarUiMsg (settings, MSG_ERR, "Cannot fork eventcmd. (%s)\n", strerror (errno));
 	} else {
 		/* parent */
 		int status;
+		PianoStation_t *songStation = NULL;
+		FILE *pipeWriteFd;
+
 		close (pipeFd[0]);
-		write (pipeFd[1], pipeBuf, strlen (pipeBuf));
+
+		pipeWriteFd = fdopen (pipeFd[1], "w");
+
+		if (curSong != NULL && stations != NULL && curStation->isQuickMix) {
+			songStation = PianoFindStationById (stations, curSong->stationId);
+		}
+
+		fprintf (pipeWriteFd,
+				"artist=%s\n"
+				"title=%s\n"
+				"album=%s\n"
+				"coverArt=%s\n"
+				"stationName=%s\n"
+				"songStationName=%s\n"
+				"pRet=%i\n"
+				"pRetStr=%s\n"
+				"wRet=%i\n"
+				"wRetStr=%s\n"
+				"songDuration=%lu\n"
+				"songPlayed=%lu\n"
+				"rating=%i\n"
+				"detailUrl=%s\n",
+				curSong == NULL ? "" : curSong->artist,
+				curSong == NULL ? "" : curSong->title,
+				curSong == NULL ? "" : curSong->album,
+				curSong == NULL ? "" : curSong->coverArt,
+				curStation == NULL ? "" : curStation->name,
+				songStation == NULL ? "" : songStation->name,
+				pRet,
+				PianoErrorToStr (pRet),
+				wRet,
+				WaitressErrorToStr (wRet),
+				player->songDuration,
+				player->songPlayed,
+				curSong == NULL ? PIANO_RATE_NONE : curSong->rating,
+				curSong == NULL ? "" : curSong->detailUrl
+				);
 
 		if (stations != NULL) {
 			/* send station list */
@@ -660,23 +863,52 @@ void BarUiStartEventCmd (const BarSettings_t *settings, const char *type,
 					settings->sortOrder);
 			assert (sortedStations != NULL);
 
-			snprintf (pipeBuf, sizeof (pipeBuf), "stationCount=%zd\n", stationCount);
-			write (pipeFd[1], pipeBuf, strlen (pipeBuf));
+			fprintf (pipeWriteFd, "stationCount=%zd\n", stationCount);
 
 			for (size_t i = 0; i < stationCount; i++) {
 				const PianoStation_t *currStation = sortedStations[i];
-				snprintf (pipeBuf, sizeof (pipeBuf), "station%zd=%s\n", i,
+				fprintf (pipeWriteFd, "station%zd=%s\n", i,
 						currStation->name);
-				write (pipeFd[1], pipeBuf, strlen (pipeBuf));
 			}
 			free (sortedStations);
 		} else {
-			const char *msg = "stationCount=0\n";
-			write (pipeFd[1], msg, strlen (msg));
+			const char * const msg = "stationCount=0\n";
+			fwrite (msg, sizeof (*msg), strlen (msg), pipeWriteFd);
 		}
 	
-		close (pipeFd[1]);
+		/* closes pipeFd[1] as well */
+		fclose (pipeWriteFd);
 		/* wait to get rid of the zombie */
 		waitpid (chld, &status, 0);
 	}
 }
+
+/*	prepend song to history, must not be a list of songs as ->next is modified!
+ */
+void BarUiHistoryPrepend (BarApp_t *app, PianoSong_t *song) {
+	if (app->settings.history != 0) {
+		PianoSong_t *tmpSong;
+
+		song->next = app->songHistory;
+		app->songHistory = song;
+
+		/* limit history's length */
+		/* start with 1, so we're stopping at n-1 and have the
+		 * chance to set ->next = NULL */
+		unsigned int i = 1;
+		tmpSong = app->songHistory;
+		while (i < app->settings.history && tmpSong != NULL) {
+			tmpSong = tmpSong->next;
+			++i;
+		}
+		/* if too many songs in history... */
+		if (tmpSong != NULL) {
+			PianoSong_t *delSong = tmpSong->next;
+			tmpSong->next = NULL;
+			if (delSong != NULL) {
+				PianoDestroyPlaylist (delSong);
+			}
+		}
+	}
+}
+
